#!/usr/bin/python

"""
Extend text_db with a CAS# validator, a command line interface, and document generation.
"""

from text_db import *
import re
import os
import os.path
import time

def valid_CASno(cas_string, debug=False):
    """
    Check N..NN-NN-N format, and the checksum digit for valid CAS number structure.
    see http://www.cas.org/expertise/cascontent/registry/checkdig.html
    for N_n .. N_4 N_3 - N_2 N_1 - R
    R = remainder([sum_{i=1}^n i N_i  ] / 10 )
    Ignores 'na' and '+secret-non-hazardous'
    >>> valid_CASno('107-07-3')
    True
    >>> valid_CASno('107-08-3')
    False
    >>> valid_CASno('107-083')
    False
    """
    for string in ['na', '+secret-non-hazardous']:
        # the first marks a non-existent CAS#
        # the last marks items with secret, non-hazardous ingredients for which we have no CAS#
        if cas_string == string:
            return True
    # check format,  
    # ^ matches the start of the string
    # \Z matches the end of the string
    regexp = re.compile('^[0-9]{2,}[-][0-9]{2}[-][0-9]\Z')
    if regexp.match(cas_string) == None:
        if debug : print >> stderr, "invalid CAS# format: '%s'" % cas_string
        return False
    # generate check digit
    casdigs = "".join(cas_string.split('-')) # remove '-'
    sumdigs = list(casdigs[:-1])
    sumdigs.reverse()
    sum=0
    for i in range(len(sumdigs)) :
        sum += (i+1)*int(sumdigs[i])
    check = sum % 10
    if int(casdigs[-1]) == check :
        return True
    else :
        if debug : print >> stderr, "invalid CAS# check: '%s' (expected %d)" % (cas_string, check)
        return False

class MSDS_manager (object) :
    """
    Manage Material Saftey Data Sheets (MSDSs)
    """
    def __init__(self, db, dir="./MSDS/") :
        self.db = db
        self.dir = dir
        self.MIMEs = ['application/pdf',
                      'text/html',
                      'text/plain']
        self.MIME_exts = ['pdf', 'html', 'txt']
        self.check_dir()
    def check_dir(self) :
        "Create the MSDS directory if it's missing"
        if os.path.isdir(self.dir) :
            return # all set to go
        elif os.path.exists(self.dir) :
            raise Exception, "Error: a non-directory file exists at %s" % self.dir
        else :
            os.mkdir(self.dir)
    def basepath(self, id) :
        assert type(id) == type(1), 'id must be an integer, not %s (%s)' \
                                    % (type(id), str(id))
        return os.path.join(self.dir, "%d" % id)
    def local_basepath(self, id) : # for symbolic links
        assert type(id) == type(1), 'id must be an integer, not %s (%s)' \
                                    % (type(id), str(id))
        return "./%d" % id
    def MIME_ext(self, mime) :
        assert mime in self.MIMEs, \
            "invalid MIME type '%s'\nshould be one of %s" % (mime, self.MIMEs)
        i = self.MIMEs.index(mime)
        ext = self.MIME_exts[i]
        return ext
    def path(self, id, mime) :
        return "%s.%s" % (self.basepath(id), self.MIME_ext(mime))
    def local_path(self, id, mime) :
        return "%s.%s" % (self.local_basepath(id), self.MIME_ext(mime))
    def save(self, id, filetext, mime='application/pdf') :
        "Save the binary byte string FILE to the path for ID"
        print >> file(self.path(id, mime), 'wb'), filetext,
    def link(self, id, target_id) :
        # target_id already exists, create a symlink to it for id.
        target_mime = self.get_MSDS_MIME(target_id)
        target_path = self.local_path(target_id, target_mime)
        path = self.path(id, target_mime)
        #os.link(self.path(target_id), self.path(id))   # hard link...
        os.symlink(target_path, path)                   # ... or soft link
    def has_MSDS_MIME(self, id, mime) :
        """
        >>> m = MSDS_manager(db=None)
        >>> print m.has_MSDS_type(102, 'pdf') # test on html
        False
        >>> print m.has_MSDS_type(102, 'html') # test on html
        True
        >>> print m.has_MSDS_type(6, 'pdf') # test on pdf symlink
        True
        """
        return os.path.exists(self.path(id, mime))
    def get_MSDS_path(self, id) :
        """
        >>> m = MSDS_manager(db=None)
        >>> print m.get_MSDS_path(102) # test on html
        ./MSDS/102.html
        >>> print m.get_MSDS_path(1) # test on pdf
        ./MSDS/1.pdf
        >>> print m.get_MSDS_path(6) # test on pdf symlink
        ./MSDS/6.pdf
        """
        for mime in self.MIMEs :
            if self.has_MSDS_MIME(id, mime) :
                return self.path(id, mime)
        return None
    def get_MSDS_MIME(self, id) :
        """
        >>> m = MSDS_manager(db=None)
        >>> print m.get_MSDS_MIME(102) # test on html
        text/html
        >>> print m.get_MSDS_MIME(1) # test on pdf
        application/pdf
        >>> print m.get_MSDS_MIME(6) # test on pdf symlink
        application/pdf
        """
        for mime in self.MIMEs :
            if self.has_MSDS_MIME(id, mime) :
                return mime
        return None
    def has_MSDS(self, id) :
        if self.get_MSDS_path(id) == None :
            return False
        return True
    def get_all(self, simlinks=True) :
        ret = []
        for record in self.db.records() :
            p = self.get_MSDS_path( int(record['ID']) )
            if p != None :
                if simlinks == False and os.path.islink( p ) :
                    continue # ignore the symbolic link
                ret.append({'ID':record['ID'], 'Name':record['Name']})
        return ret

class docgen (object) :
    "Generate the officially required documents"
    def __init__(self, db) :
        self.db = db
    def _latex_safe(self, string):
        string = string.replace('%', '\%')
        string = string.replace('>', '$>$')
        string = string.replace('<', '$<$')
        return string
    def _set_main_target(self, target):
        print >> file('./docs/main.tex', 'w'), \
            """\documentclass[letterpaper]{article}

\input{%s}
""" % target
    def _make_pdf(self, target_file):
        os.system('cd ./docs && make pdf')
        path = os.path.join('./docs/', target_file)
        os.system('cp ./docs/main.pdf %s' % path)
        return path
    def inventory(self, namewidth='a') :
        "Create a pdf list of all currently owned chemicals."
        pp = db_pretty_printer(self.db)
        active_ids = []
        for record in self.db.records() :
            if record['Disposed'] == '' : # get ids for chemicals we still have
                active_ids.append(record['db_id'])
        active_fields = ['ID', 'Name', 'Amount',
                         'H', 'F', 'R', 'O', 'M', 'C', 'T']
        width = {}
        for field in active_fields :
            width[field] = 'a'
        width['Name'] = namewidth
        ## Plain text method
        #string = "Chemical inventory:\t\tGenerated on %s\n\n" \
        #         % time.strftime('%Y-%m-%d')
        #string += pp.multi_record_string(active_ids, active_fields,
        #                                 width=width, FS=' ')
        # return string
        ## Latex method
        string = "\\begin{longtable}{l l l c c c c c c c}\n"
        string += ('%% The header for the remaining page(s) of the table...\n'
                   'ID & Name & Amount & H & F & R & O & M & C & T \\\\\n'
                   '\\hline\n'
                   '\\endhead\n')
        for db_id in active_ids :
            record = self.db.record(db_id)
            string += "  %s & %s & %s & %s & %s & %s & %s & %s & %s & %s \\\\\n" \
                      % (self._latex_safe(record['ID']),
                         self._latex_safe(record['Name']),
                         self._latex_safe(record['Amount']),
                         self._latex_safe(record['H']),
                         self._latex_safe(record['F']),
                         self._latex_safe(record['R']),
                         self._latex_safe(record['O']),
                         self._latex_safe(record['M']),
                         self._latex_safe(record['C']),
                         self._latex_safe(record['T']))
        string += "\\end{longtable}\n"
        print >> file('./docs/inventory_data.tex', 'w'), string
        ## alter main.tex to point to the inventory template.
        self._set_main_target('inventory_template')
        ## run latex
        path = self._make_pdf('inventory.pdf')
        return path
    def door_warning(self,
                     valid_record=lambda r: r['Disposed'] == '') :
        """create a warning NFPA diamond and list of the most dangerous
        chemicals for which valid_record(record) is true.  For
        example, to generate a door warning for the front door use
          door_warning(lambda r: r['Disposed'] == '')
        or to generate the warning for the fridge
          door_warning(lambda r: r['Location'] == 'Refrigerator')
        Note that valid_record defaults to the first example.
        """
        pp = db_pretty_printer(self.db)
        all_ids = range(self.db.len_records())

        # Search the database to find the nasties
        NFPA_maxs = {'H':0, 'F':0, 'R':0, 'O':[]}
        Mutagens = []
        Carcinogens = []
        Teratogens = []
        Healths = []
        Fires = []
        Reactivities = []
        Others = []
        for record in self.db.records() :
            if valid_record(record) :
                for field in ['H', 'F', 'R', 'O'] :
                    r = record[field]
                    if r != '' and r != '?' :
                        if field != 'O' and int(r) > NFPA_maxs[field] :
                            NFPA_maxs[field] = int(r)
                        elif field == 'O' and not r in NFPA_maxs['O'] :
                            NFPA_maxs[field].append(r)
                for field,array in zip(['M','C','T'],
                                       [Mutagens,
                                        Carcinogens,
                                        Teratogens]) :
                    if record[field] != '' and record[field] != '?':
                        array.append(record['db_id'])
        # now that we've found the max NFPAs,
        # find all the chemicals at those levels
        for record in self.db.records() :
            if valid_record(record) :
                for field,array in zip(['H', 'F', 'R', 'O'],
                                       [Healths, Fires,
                                        Reactivities, Others]) :
                    r = record[field]
                    if r != '' and r != '?' :
                        if field != 'O' and int(r) == NFPA_maxs[field] :
                            array.append(record['db_id'])
                        elif field == 'O' and r in NFPA_maxs['O'] :
                            array.append(record['db_id'])

        ## generate the output
        # first, update the NFPA grapic code
        if 'OX' in NFPA_maxs['O'] : OX = 'y'
        else :                      OX = 'n'
        if 'W'  in NFPA_maxs['O'] : W  = 'y'
        else :                      W  = 'n'
        os.system('./docs/mp/gen_NFPA.sh %d %d %d %s %s > ./docs/mp/NFPA.mp'
                  % (NFPA_maxs['H'], NFPA_maxs['F'], NFPA_maxs['R'], OX, W))
        # now generate a list of the nasties ( Amount & ID & Name )
        string = "\\begin{tabular}{r r l}\n"
        for field,name,array in zip(['H', 'F', 'R', 'O'],
                                    ['Health', 'Fire',
                                     'Reactivity', 'Other'],
                                    [Healths, Fires,
                                     Reactivities, Others]) :
            if (not hasattr(NFPA_maxs[field], '__len__')) \
                    or len(NFPA_maxs[field]) > 0 :
                string += "  \multicolumn{3}{c}{\Tstrut %s : %s} \\\\\n" \
                    % (name, NFPA_maxs[field])
            else : # Print "Other" instead of "Other : []"
                string += "  \multicolumn{3}{c}{\Tstrut %s} \\\\\n" \
                    % (name)
            for db_id in array :
                record = self.db.record(db_id)
                string += "  %s  &  %s &  %s \\\\\n" \
                    % (self._latex_safe(record['Amount']),
                       self._latex_safe(record['ID']),
                       self._latex_safe(record['Name']))
            if len(array) == 0 :
                string += "  \multicolumn{3}{c}{ --- } \\\\\n"
        for hazard,array in zip(['Mutagens','Carcinogens','Teratogens'],
                                [Mutagens, Carcinogens, Teratogens]) :
            string += "  \multicolumn{3}{c}{\Tstrut %s} \\\\\n" % (hazard)
            for db_id in array :
                record = self.db.record(db_id)
                string += "  %s  &  %s &  %s \\\\\n" \
                    % (self._latex_safe(record['Amount']),
                       self._latex_safe(record['ID']),
                       self._latex_safe(record['Name']))
            if len(array) == 0 :
                string += "  \multicolumn{3}{c}{ --- } \\\\\n"                
        string += "\\end{tabular}\n"
        print >> file('./docs/door_data.tex', 'w'), string
        ## alter main.tex to point to the door template.
        self._set_main_target('door_template')
        ## run latex
        path = self._make_pdf('door_warning.pdf')
        return path

def _test():
    import doctest
    doctest.testmod()

def open_IOfiles(ifilename=None, ofilename=None, debug=False):
    if ifilename :
        if debug :  print >> stderr, "open input file '%s'" % ifilename
        ifile = file(ifilename, 'r')
    else :
        ifile = stdin
    if ofilename :
        if debug :  print >> stderr, "open output file '%s'" % ofilename
        ofile = file(ofilename, 'w')
    else :
        ofile = stdout    
    return (ifile, ofile)

def close_IOfiles(ifilename=None, ifile=stdin, 
                  ofilename=None, ofile=stdout,
                  debug=False):
    if ifilename :
        if debug :  print >> stderr, "close input file '%s'" % ifilename
        ifile.close()
    if ofilename :
        if debug :  print >> stderr, "close output file '%s'" % ofilename
        ofile.close()


if __name__ == "__main__" :
    from optparse import OptionParser

    parser = OptionParser(usage="usage: %prog [options]", version="%prog 0.1")

    parser.add_option('-f', '--input-file', dest="ifilename",
                      help="Read input from FILE (default stdin)",
                      type='string', metavar="FILE")
    parser.add_option('-o', '--output-file', dest="ofilename",
                      help="Write output to FILE (default stdout)",
                      type='string', metavar="FILE")
    parser.add_option('-d', '--delimiter', dest="FS", # field seperator
                      help="Set field delimiter (default '%default')",
                      type='string', metavar="DELIM", default='\t')
    parser.add_option('-p', '--print-fields', dest="print_fields",
                      help="Only print certain fields (e.g. 0,3,4,2)",
                      type='string', metavar="FIELDS")
    parser.add_option('-r', '--print-records', dest="print_records",
                      help="Only print certain records (e.g. 0:3)",
                      type='string', metavar="RECORDS")
    parser.add_option('-w', '--column-width', dest="width",
                      help="Set column width for short-format output.",
                      type='string', metavar="WIDTH")
    parser.add_option('-L', '--long-format', dest="long_format",
                      help="Print long format (several lines per record)",
                      action='store_true', default=False)
    parser.add_option('-l', '--short-format', dest="long_format",
                      help="Print short format (default) (one lines per record)",
                      action='store_false', default=False)
    parser.add_option('-t', '--test', dest="test",
                      help="Run docutils tests on db.py",
                      action="store_true", default=False)
    parser.add_option('-V', '--validate', dest="validate",
                      help="Validate CAS#s (no other output)",
                      action="store_true", default=False)
    parser.add_option('-v', '--verbose', dest="verbose",
                      help="Print lots of debugging information",
                      action="store_true", default=False)

    (options, args) = parser.parse_args()
    parser.destroy()

    ifile,ofile = open_IOfiles(options.ifilename, options.ofilename,
                               options.verbose)

    if options.test :
        _test()
    elif options.validate :
        db = text_db(filename=None)
        pp = db_pretty_printer(db)

        # read in and parse the file
        db._parse(ifile.read())

        CAS_DELIM = ',' # seperate CAS entries for chemicals with multiple CAS numbers
        PERCENT_DELIM = ':' # seperate CAS number from ingredient percentage
        for record in db.records() :
            valid = True
            cas = record['CAS#']
            if len(cas.split(CAS_DELIM)) == 0 : # cas = 'N...N-NN-N'
                if not valid_CASno(cas, options.verbose) :
                    valid = False
                    print >> ofile, "Invalid CAS# in record: '%s'" % cas
            else : # cas = 'N...N-NN-N:X%,N...N-NN-N:Y%,...'
                for casterm in cas.split(CAS_DELIM) : # casterm = 'N...N-NN-N:X%'
                    c = casterm.split(PERCENT_DELIM)[0]   # c = 'N...N-NN-N'
                    if not valid_CASno(c, options.verbose) :
                        valid = False
                        print >> ofile, "Invalid CAS* in record: '%s'" % c
            if not valid :
                print >> ofile, "in record %s: %s" % (record['ID'], record['Name'])
                #pp.full_record_string(record)

    else :
        db = text_db(filename=None)

        # read in and parse the file
        db._parse(ifile.read())
        pp = db_pretty_printer(db)
        if options.long_format :
            for id in pp._norm_record_ids(options.print_records) :
                string = pp.full_record_string_id(id)
        else :
            # pythonize the width option
            if (options.width != None
                and options.width != 'a'
                and len(options.width.split(':')) == 1
                ) :
                width = int(options.width)
            elif len(options.width.split(':')) > 1 :
                width = {}
                for kv in options.width.split(',') :
                    spl = kv.split(':')
                    assert len(spl) == 2, 'invalid width "%s" in "%s"' % (kv, options.width)
                    if spl[1] == 'a' :
                        width[spl[0]] = spl[1]
                    else :
                        width[spl[0]] = int(spl[1])

            string = pp.multi_record_string(options.print_records,
                                            options.print_fields,
                                            width,
                                            options.FS)
            print >> ofile, string,
            
    close_IOfiles(options.ifilename, ifile,
                  options.ofilename, ofile, options.verbose)
